/*
 * SIP-tarpit
 * Listens on 5060 udp by default
 *
 */
mod sip;
mod config;
mod database;

use crate::sip::generic::*;
use crate::sip::not_found::*;
use crate::sip::options::*;
use crate::sip::register::*;
use crate::sip::unauthorized::*;
use crate::database::*;
use rand::Rng;
use std::error::Error;
use std::net::SocketAddr;
use std::io;
use tokio::net::UdpSocket;
use tokio::time::{sleep, Duration};
use log::LevelFilter;
use log4rs::append::file::FileAppender;
use log4rs::append::console::ConsoleAppender;
use log4rs::config::{Appender, Config, Root};
use log4rs::encode::pattern::PatternEncoder;
use rusqlite::Connection;
use std::process::Command;

struct Server {
    socket: UdpSocket,
    buf: Vec<u8>,
    to_send: Option<(usize, SocketAddr)>,
    config: config::Config,
    db_con: Connection
}

impl Server {
    async fn run(self) -> Result<(), io::Error> {
        let Server {
            socket,
            mut buf,
            mut to_send,
            config,
            db_con
        } = self;

        loop {
            // First we check to see if there's a message we need to answer.
            // If so then we try to send it back to the original source, waiting
            // until it's writable and we're able to do so.
            if let Some((size, peer)) = to_send {
                // save peer
                let str_peer_ip = peer.ip().to_string();
                match get_suspicious_by_ip_addr(&db_con, str_peer_ip.as_str()) {
                    Ok(_) => {
                        // was added to database, do nothing
                        // log::info!("Peer already added to database");
                    },
                    Err(_e) => {
                        // was not in database
                        // log::info!("Error getting suspicious peer from database: {}", _e);
                        log::info!("Suspicious peer: {}", peer.ip());
                        match add_suspicious(&db_con, str_peer_ip.as_str()) {
                            Ok(_) => {
                                // launch action script
                                let output = Command::new(config.action_script.to_owned())
                                    .arg(str_peer_ip.as_str())
                                    .output();
                                match output {
                                    Ok(_) => log::info!("Action script executed"),
                                    Err(e) => log::info!("Error executing action script {}", e)
                                }
                            },
                            Err(_) => return Err(io::Error::new(io::ErrorKind::Other, "Error adding suspicious peer to database")),
                        }
                    }
                }

                // test type of message received
                let msg: String;
                let msg_result = String::from_utf8(buf[..size].to_vec());

                match msg_result {
                    Ok(m) => msg = m,
                    Err(_) => {
                        msg = "".to_string();
                        log::warn!("Error converting message to string");
                    }
                }

                match msg {
                    msg if msg.contains("OPTIONS") => {
                        // esperar - tarpit!
                        sleep(Duration::from_millis(config.delay)).await;

                        let options = Options::parse(&msg);
                        // respondemos con un 404 not found
                        // o campo via non é como vai... pero ñapa
                        let not_found = NotFoundPacket {
                            command: "SIP/2.0 404 Not Found".to_string(),
                            via: options.via,
                            call_id: options.call_id,
                            from: options.from,
                            to: options.to,
                            cseq: options.cseq,
                            accept: Some("application/sdp".to_string()),
                            allow: Some("OPTIONS, REGISTER, SUBSCRIBE, NOTIFY, PUBLISH, INVITE, ACK, BYE, CANCEL, UPDATE, PRACK, MESSAGE, REFER".to_string()),
                            supported: Some("100rel, timer, replaces, norefersub".to_string()),
                            accept_encoding: Some("itentity".to_string()),
                            accept_language: Some("en".to_string()),
                            server: Some("Asterisk PBX 18.9.0".to_string()),
                            content_length: Some(0)
                        };
                        let amt = socket.send_to(&not_found.serialize(), &peer).await?;
                        log::debug!("Sent {}/{} bytes to {}",
                            amt,
                            &not_found.serialize().len(),
                            peer
                        );
                    },
                    msg if msg.contains("REGISTER") => {
                        // esperar - tarpit!
                        sleep(Duration::from_millis(config.delay)).await;

                        let register = Register::parse(&msg);

                        // responde aleatoriamente unauthorized (a ext. existe)
                        // ou un notfound genérico (non existe)
                        let mut num = rand::thread_rng().gen_range(0..1000);

                        //To: "6666666666"<sip:6666666666@xinu.me>
                        // get exten, if is not in range send forbidden
                        let mut exten = String::new();
                        if register.to.contains("<") && register.to.contains(">") {
                            let start = register.to.find("<").unwrap_or(0); //ou inicio
                            let end = register.to.find(">").unwrap_or(register.to.len()); //ou fin
                            let ext_with_host = &register.to[start..end];
                            exten = ext_with_host
                                .split("@")
                                .collect::<Vec<&str>>()[0]
                                .split(":")
                                .collect::<Vec<&str>>()[1]
                                .to_string();
                        }
                        // si ten máis de 5 díxitos é a exten de proba que manda
                        // o sipvicius como test (normalmente é a primeira)
                        if exten.len() > 5 {
                            num = 9000;
                        }

                        if num < 500 {
                            let unauthorized = Unauthorized {
                                command: "SIP/2.0 401 Unauthorized".to_string(),
                                via: register.via,
                                call_id: register.call_id,
                                from: register.from,
                                to: format!("{};tag={}", register.to, register.branch),
                                cseq: register.cseq,
                                www_authenticate: Some("Digest realm=\"asterisk\",nonce=\"1670408728/f2288b61ff3e1c5a8fbad2282734a53b\",opaque=\"37899594559ff83d\",algorithm=md5,qop=\"auth\"".to_string()),
                                server: Some("Asterisk PBX 18.9.0".to_string()),
                                content_length: Some(0)
                            };
                            // println!("{}", std::str::from_utf8(&unauthorized.serialize()).unwrap());
                            let amt = socket.send_to(&unauthorized.serialize(), &peer).await?;
                            log::debug!(
                                "Unauthorized: Sent {}/{} bytes to {}",
                                amt,
                                &unauthorized.serialize().len(),
                                peer
                                );
                        } else {
                            let notfound = Generic {
                                command: "SIP/2.0 404 NotFound".to_string(),
                                via: register.via,
                                call_id: register.call_id.unwrap(),
                                from: register.from,
                                to: format!("{};tag={}", register.to, register.branch),
                                cseq: register.cseq,
                                max_forwards: 70,
                                server: Some("Asterisk PBX 18.9.0".to_string()),
                                content_length: Some(0)
                            };
                            // println!("{}", std::str::from_utf8(&notfound.serialize()).unwrap());
                            let amt = socket.send_to(&notfound.serialize(), &peer).await?;
                            log::debug!(
                                "NotFound: Sent {}/{} bytes to {}",
                                amt,
                                &notfound.serialize().len(),
                                peer
                                );
                        }
                    },
                    _ => println!("Packet don't known, yet..."),
                }
            }

            // If we're here then `to_send` is `None`, so we take a look for the
            // next message we're going to answer.
            to_send = Some(socket.recv_from(&mut buf).await?);
        }
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn Error>> {
    let config: config::Config = config::Config::new();

    let log_level_filter = match config.log_level.to_lowercase().as_str() {
        "trace" => LevelFilter::Trace,
        "debug" => LevelFilter::Debug,
        "info" => LevelFilter::Info,
        "warn" => LevelFilter::Warn,
        "error" => LevelFilter::Error,
        _ => LevelFilter::Info,
    };

    if log_level_filter == LevelFilter::Debug {
        log::info!("Debug mode enabled");
        println!("{:#?}", config);
    }

    let logconfig: log4rs::Config;

    // write log to stdout or to file
    if config.log_file.to_lowercase().as_str() == "console" {
        let stdout = ConsoleAppender::builder()
            .encoder(Box::new(PatternEncoder::new("{d} - {l} - {m}\n")))
            .build();

        logconfig = Config::builder()
            .appender(Appender::builder().build("stdout", Box::new(stdout)))
            .build(Root::builder().appender("stdout").build(log_level_filter))
            .unwrap();
    } else {

        let logfile = FileAppender::builder()
            .encoder(Box::new(PatternEncoder::new("{d} - {l} - {m}\n")))
            .build(config.log_file.as_str())
            .unwrap();

        logconfig = Config::builder()
            .appender(Appender::builder().build("logfile", Box::new(logfile)))
            .build(Root::builder().appender("logfile").build(log_level_filter))
            .unwrap();
    }

    log4rs::init_config(logconfig).unwrap();

    let db_con = init_db(config.database_location.as_str())?;

    let addr = format!("{}:{}", config.bind_addr, config.bind_port);
    let socket = UdpSocket::bind(&addr).await?;
    log::info!("Listening on: {}", socket.local_addr()?);

    let server = Server {
        socket,
        buf: vec![0; 1024],
        to_send: None,
        config,
        db_con
    };

    // This starts the server task.
    server.run().await?;

    Ok(())
}
